@@ -74,42 +74,16 @@ static int open_device(const char *device)
 	return fd;
 }
 
-static int ioctl_device(int fd, int req, void *arg)
-{
-	if (fd < 0)
-		return -1;
 
-	if (ioctl(fd, req, arg)) {
-		char *msg = "UNKNOWN";
-
-		switch (req) {
-		case PLOOP_IOC_BALLOON:
-			msg = "PLOOP_IOC_BALLOON";
-			break;
-		case PLOOP_IOC_FREEBLKS:
-			msg = "PLOOP_IOC_FREEBLKS";
-			break;
-		case PLOOP_IOC_RELOCBLKS:
-			msg = "PLOOP_IOC_RELOCBLKS";
-			break;
-		case PLOOP_IOC_FBGET:
-			msg = "PLOOP_IOC_FBGET";
-			break;
-		case PLOOP_IOC_DISCARD_INIT:
-			msg = "PLOOP_IOC_DISCARD_INIT";
-			break;
-		case PLOOP_IOC_DISCARD_WAIT:
-			msg = "PLOOP_IOC_DISCARD_WAIT";
-			break;
-		case PLOOP_IOC_DISCARD_FINI:
-			msg = "PLOOP_IOC_DISCARD_FINI";
-			break;
-		}
-		ploop_err(errno, "%s", msg);
-		return(SYSEXIT_DEVIOC);
-	}
-	return 0;
-}
+#define ioctl_device(fd, req, arg)					\
+	({								\
+		int __ret = 0;						\
+		if (ioctl(fd, req, arg)) {				\
+			ploop_err(errno, "Error in ioctl(" #req ")");	\
+			__ret = SYSEXIT_DEVIOC;				\
+		}							\
+		__ret;							\
+	 })
 
 static int fsync_balloon(int fd)
 {
@@ -845,35 +819,119 @@ err:
 	return ret;
 }
 
-static int ploop_trim(const char *mount_point)
+static int trim_stop = 0;
+static void stop_trim_handler(int sig)
 {
-	struct fstrim_range range = {0, ULLONG_MAX, PAGE_SIZE};
+	trim_stop = 1;
+}
+
+static int ploop_trim(const char *mount_point, __u64 minlen_b)
+{
+	struct fstrim_range range = {0, ULLONG_MAX, minlen_b};
 	int fd, ret;
 
+	struct sigaction sa = {
+		.sa_handler     = stop_trim_handler,
+	};
+	sigemptyset(&sa.sa_mask);
+
+        if (sigaction(SIGUSR1, &sa, NULL)) {
+                ploop_err(errno, "Can't set signal handler");
+                exit(1);
+        }
+
 	fd = open(mount_point, O_RDONLY);
 	if (fd < 0) {
 		ploop_err(errno, "Can't open mount_point");
 		return -1;
 	}
 
+	sys_syncfs(fd);
+
 	ret = ioctl(fd, FITRIM, &range);
-	if (ret < 0)
-		ploop_err(errno, "Can't trim file system");
+	if (ret < 0) {
+		if (trim_stop)
+			ret = 0;
+		else
+			ploop_err(errno, "Can't trim file system");
+	}
 
 	close(fd);
 
 	return ret;
 }
 
-int ploop_discard(const char *device, const char *mount_point)
-{
-	int fd, ret, status;
-	int exit_code = 0;
-	pid_t tpid;
+#define ilog2(n) (			\
+	(n) >= (1 << 9) ? 9 :		\
+	(n) & 1 << 8	 ? 8 :		\
+	(n) & 1 << 7	 ? 7 :		\
+	(n) & 1 << 6	 ? 6 :		\
+	(n) & 1 << 5	 ? 5 :		\
+	(n) & 1 << 4	 ? 4 :		\
+	(n) & 1 << 3	 ? 3 :		\
+	(n) & 1 << 2	 ? 2 :		\
+	(n) & 1 << 1	 ? 1 :	0	\
+)
 
-	fd = open_device(device);
-	if (fd == -1)
-		return SYSEXIT_OPEN;
+#define DISCARD_TABLE_SIZE 10
+
+static int discard_collect_stat(int fd, __u32 *distrib)
+{
+	struct ploop_freeblks_ctl  *freeblks = NULL;
+	int ret, i;
+
+	ret = freeblks_alloc(&freeblks, 0);
+	if (ret)
+		return ret;
+
+	ret = ioctl_device(fd, PLOOP_IOC_FBGET, freeblks);
+	if (ret)
+		goto free;
+
+	if (freeblks->n_extents == 0) {
+		// This should never happen, and we don't want
+		// to return an error here.
+		ploop_err(EINVAL, "The number of extents is zero");
+		goto free;
+	}
+
+	ret = freeblks_alloc(&freeblks, freeblks->n_extents);
+	if (ret)
+		goto free;
+
+	ret = ioctl_device(fd, PLOOP_IOC_FBGET, freeblks);
+	if (ret)
+		goto free;
+
+	for (i = 0; i < freeblks->n_extents; i++) {
+		__u32 len = freeblks->extents[i].len;
+		distrib[ilog2(len)] += len;
+	}
+	ret = ioctl_device(fd, PLOOP_IOC_FBDROP, 0);
+	if (ret)
+		goto free;
+free:
+	free(freeblks);
+	return ret;
+}
+
+enum {
+	PLOOP_DISCARD_STAT,
+	PLOOP_DISCARD_COMPACT,
+
+	PLOOP_DISCARD_MAX
+};
+
+static int __ploop_discard(int fd, const char *device, const char *mount_point,
+			int state, __u32 *minlen_c, __u32 cluster, __u32 to_free)
+{
+	pid_t tpid;
+	int exit_code = 0, ret, status;
+	__u32 distrib[DISCARD_TABLE_SIZE], size = 0;
+
+	memset(distrib, 0, sizeof(distrib));
+
+	ploop_log(3, "Trying to find free extents bigger than %u clusters", *minlen_c);
 
 	ret = ioctl_device(fd, PLOOP_IOC_DISCARD_INIT, NULL);
 	if (ret) {
@@ -893,7 +951,7 @@ int ploop_discard(const char *device, const char *mount_point)
 	}
 
 	if (tpid == 0) {
-		ret = ploop_trim(mount_point);
+		ret = ploop_trim(mount_point, (__u64) *minlen_c * cluster);
 		if (ret < 0)
 			exit_code = 1;
 
@@ -908,23 +966,60 @@ int ploop_discard(const char *device, const char *mount_point)
 		struct ploop_balloon_ctl b_ctl;
 
 		ploop_log(0, "Waiting");
-		ret = ioctl_device(fd, PLOOP_IOC_DISCARD_WAIT, NULL);
-		if (ret) {
+		ret = ioctl(fd, PLOOP_IOC_DISCARD_WAIT, NULL);
+		if (ret < 0) {
 			ploop_err(errno, "Waiting for a discard request failed");
 			break;
+		} else if (ret == 0)
+			break;
+
+		ret = ioctl(fd, PLOOP_IOC_FBFILTER, *minlen_c);
+		if (ret < 0) {
+			ploop_err(errno, "Can't filter free blocks");
+			break;
+		} else if (ret == 0) {
+			/* Nothing to do */
+			ret = ioctl_device(fd, PLOOP_IOC_FBDROP, 0);
+			if (ret)
+				break;
+			continue;
+		} else
+			size += ret;
+
+		switch (state) {
+		case PLOOP_DISCARD_COMPACT:
+			memset(&b_ctl, 0, sizeof(b_ctl));
+			b_ctl.keep_intact = 1;
+			ret = ioctl_device(fd, PLOOP_IOC_BALLOON, &b_ctl);
+			if (ret)
+				break;
+
+			if (b_ctl.mntn_type == PLOOP_MNTN_OFF) {
+				ploop_log(0, "Unexpected maintenance type 0x%x", b_ctl.mntn_type);
+				ret = -1;
+				break;
+			}
+
+			if (size >= to_free) {
+				ploop_log(3, "Killing the trim process %d", tpid);
+				kill(tpid, SIGUSR1);
+				ret = ioctl_device(fd, PLOOP_IOC_DISCARD_FINI, NULL);
+				if (ret < 0 && errno != EBUSY)
+					ploop_err(errno, "Can't finalize a discard mode");
+			}
+
+			ploop_log(0, "Starting relocation");
+			ret = ploop_balloon_relocation(fd, &b_ctl, device);
+			break;
+		case PLOOP_DISCARD_STAT:
+			ploop_log(0, "Getting extents");
+			ret = discard_collect_stat(fd, distrib);
+			break;
+		default:
+			ret = -EINVAL;
+			break;
 		}
 
-		memset(&b_ctl, 0, sizeof(b_ctl));
-		b_ctl.keep_intact = 1;
-		ret = ioctl_device(fd, PLOOP_IOC_BALLOON, &b_ctl);
-		if (ret)
-			break;
-
-		if (b_ctl.mntn_type == PLOOP_MNTN_OFF)
-			break;
-
-		ploop_log(0, "Start relocation");
-		ret = ploop_balloon_relocation(fd, &b_ctl, device);
 		if (ret)
 			break;
 	}
@@ -939,8 +1034,6 @@ int ploop_discard(const char *device, const char *mount_point)
 		kill(tpid, SIGKILL);
 	}
 
-	close(fd);
-
 	ret = waitpid(tpid, &status, 0);
 	if (ret == -1) {
 		ploop_err(errno, "wait() failed");
@@ -955,5 +1048,62 @@ int ploop_discard(const char *device, const char *mount_point)
 		exit_code = 1;
 	}
 
+	if (exit_code == 0 && state == PLOOP_DISCARD_STAT) {
+		int j;
+
+		for (j = DISCARD_TABLE_SIZE - 1; j >= 0; j--)
+			ploop_log(3, "%10d\t%u", j, distrib[j]);
+
+		for (j = DISCARD_TABLE_SIZE - 1; j >= 0; j--) {
+			if (to_free <= distrib[j]) {
+				*minlen_c = (1 << j);
+				break;
+			}
+
+			to_free -= distrib[j];
+		}
+	}
+
 	return exit_code;
 }
+
+int ploop_discard(const char *device, const char *mount_point,
+				__u64 minlen_b, __u64 to_free)
+{
+	int fd, ret, state;
+	int blocksize;
+	__u32  cluster, minlen_c;
+
+	if (ploop_get_attr(device, "block_size", &blocksize)) {
+		ploop_err(0, "Can't find block size");
+		return SYSEXIT_SYSFS;
+	}
+	cluster = S2B(blocksize);
+
+	if (minlen_b || to_free == ~0ULL)
+		state = PLOOP_DISCARD_COMPACT;
+	else
+		state = PLOOP_DISCARD_STAT;
+
+	minlen_c = (minlen_b + cluster - 1) / cluster;
+	to_free = to_free  / cluster;
+	if (!to_free) {
+		ploop_err(0, "Can't shrink by less than %d bytes", cluster);
+		return 0;
+	}
+
+	fd = open_device(device);
+	if (fd == -1)
+		return SYSEXIT_OPEN;
+
+	for (; state < PLOOP_DISCARD_MAX; state++) {
+		ret = __ploop_discard(fd, device, mount_point, state,
+					&minlen_c, cluster, to_free);
+		if (ret)
+			break;
+	}
+
+	close(fd);
+
+	return ret;
+}
