@@ -29,80 +29,205 @@
 #include <malloc.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
-#include <getopt.h>
 #include <linux/types.h>
 #include <string.h>
 
 #include "ploop.h"
 
-#define CLUSTER DEF_CLUSTER
-
-static int send_buf(int ofd, void *iobuf, int len, off_t pos)
+static int nwrite(int fd, const void *buf, int len)
 {
-	int n;
+	while (len) {
+		int n;
+
+		n = write(fd, buf, len);
+		if (n < 0) {
+			if (errno == EINTR)
+				continue;
+			return -1;
+		}
+		if (n == 0)
+			break;
+		len -= n;
+		buf += n;
+	}
 
 	if (len == 0)
 		return 0;
 
-	n = pwrite(ofd, iobuf, len, pos);
-	if (n < 0)
-		return n;
-	if (n != len) {
-		errno = EIO;
-		return -1;
-	}
+	errno = EIO;
+	return -1;
+}
+
+static int send_buf(int ofd, void *iobuf, int len, off_t pos)
+{
+	struct xfer_desc desc = { .marker = PLOOPCOPY_MARKER };
+
+	/* Header */
+	desc.size = len;
+	desc.pos = pos;
+	if (nwrite(ofd, &desc, sizeof(desc)))
+		return SYSEXIT_WRITE;
+
+	/* Data */
+	if (len && nwrite(ofd, iobuf, len))
+		return SYSEXIT_WRITE;
+
 	return 0;
 }
 
-static int get_image_info(const char *device, char **send_from_p, char **format_p)
+static int nread(int fd, void * buf, int len)
 {
-	FILE *fp;
-	int len;
-	char nbuf[4096];
+	while (len) {
+		int n;
 
-	if (memcmp(device, "/dev/", 5) == 0)
-		device += 5;
+		n = read(fd, buf, len);
+		if (n < 0) {
+			if (errno == EINTR)
+				continue;
+			return -1;
+		}
+		if (n == 0)
+			break;
+		len -= n;
+		buf += n;
+	}
 
-	snprintf(nbuf, sizeof(nbuf)-1, "/sys/block/%s/pdelta/0/image", device);
-	fp = fopen(nbuf, "r");
-	if (fp == NULL) {
-		ploop_err(errno, "fopen sysfs image");
-		return -1;
-	}
-	if (fgets(nbuf, sizeof(nbuf), fp) == NULL) {
-		ploop_err(errno, "read sysfs image");
-		fclose(fp);
-		return -1;
-	}
-	len = strlen(nbuf);
-	if (len > 0 && nbuf[len-1] == '\n') {
-		len--;
-		nbuf[len] = 0;
-	}
-	*send_from_p = strdup(nbuf);
-	fclose(fp);
+	if (len == 0)
+		return 0;
 
-	snprintf(nbuf, sizeof(nbuf)-1, "/sys/block/%s/pdelta/0/format", device);
-	fp = fopen(nbuf, "r");
-	if (fp == NULL) {
-		ploop_err(errno, "fopen sysfs format");
-		return -1;
+	errno = EIO;
+	return -1;
+}
+
+int receive_process(const char *dst)
+{
+	int ofd, ret;
+	__u64 cluster = 0;
+	void *iobuf = NULL;
+
+	if (isatty(0) || errno == EBADF) {
+		ploop_err(errno, "Invalid input stream: must be pipelined "
+				"to a pipe or a socket");
+		return SYSEXIT_PARAM;
 	}
-	if (fgets(nbuf, sizeof(nbuf), fp) == NULL) {
-		ploop_err(errno, "read sysfs format");
-		fclose(fp);
-		return -1;
+
+	ofd = open(dst, O_WRONLY|O_CREAT|O_EXCL, 0600);
+	if (ofd < 0) {
+		ploop_err(errno, "Can't open %s", dst);
+		return SYSEXIT_CREAT;
 	}
-	len = strlen(nbuf);
-	if (len > 0 && nbuf[len-1] == '\n') {
-		len--;
-		nbuf[len] = 0;
+
+	/* Read data */
+	for (;;) {
+		int n;
+		struct xfer_desc desc;
+
+		if (nread(0, &desc, sizeof(desc)) < 0) {
+			ploop_err(0, "Error in nread(desc)");
+			ret = SYSEXIT_READ;
+			goto out;
+		}
+		if (desc.marker != PLOOPCOPY_MARKER) {
+			ploop_err(0, "Stream corrupted");
+			ret = SYSEXIT_PROTOCOL;
+			goto out;
+		}
+		if (desc.size > cluster) {
+			free(iobuf);
+			iobuf = NULL;
+			cluster = desc.size;
+			if (posix_memalign(&iobuf, 4096, cluster)) {
+				ploop_err(errno, "posix_memalign");
+				ret = SYSEXIT_MALLOC;
+				goto out;
+			}
+		}
+		if (desc.size == 0)
+			break;
+
+		if (nread(0, iobuf, desc.size)) {
+			ploop_err(errno, "Error in nread data");
+			ret = SYSEXIT_READ;
+			goto out;
+		}
+		n = pwrite(ofd, iobuf, desc.size, desc.pos);
+		if (n != desc.size) {
+			if (n < 0)
+				ploop_err(errno, "Error in pwrite");
+			else
+				ploop_err(0, "Error: short pwrite");
+			ret = SYSEXIT_WRITE;
+			goto out;
+		}
 	}
-	*format_p = strdup(nbuf);
-	fclose(fp);
+
+	if (fsync(ofd)) {
+		ploop_err(errno, "Error in fsync");
+		ret = SYSEXIT_WRITE;
+		goto out;
+	}
+	ret = 0;
+
+out:
+	if (close(ofd)) {
+		ploop_err(errno, "Error in close");
+		if (!ret)
+			ret = SYSEXIT_WRITE;
+	}
+	if (ret)
+		unlink(dst);
+	free(iobuf);
+
+	return ret;
+}
+
+static int get_image_info(const char *device, char **send_from_p,
+		char **format_p, int *blocksize)
+{
+	int top_level;
+
+	if (ploop_get_attr(device, "top", &top_level)) {
+		ploop_err(0, "Can't find top delta");
+		return SYSEXIT_SYSFS;
+	}
+
+	if (ploop_get_attr(device, "block_size", blocksize)) {
+		ploop_err(0, "Can't find block size");
+		return SYSEXIT_SYSFS;
+	}
+
+	if (find_delta_names(device, top_level, top_level,
+				send_from_p, format_p)) {
+		ploop_err(errno, "find_delta_names");
+		return SYSEXIT_SYSFS;
+	}
+
 	return 0;
 }
 
+static int run_cmd(const char *cmd)
+{
+	int st;
+
+	if (!cmd)
+		return 0;
+
+	st = system(cmd);
+	if (!st)
+		return 0;
+
+	if (st == -1)
+		ploop_err(errno, "Can't execute %s", cmd);
+	else if (WIFEXITED(st))
+		ploop_err(0, "Command %s failed with code %d", cmd, WEXITSTATUS(st));
+	else if (WIFSIGNALED(st))
+		ploop_err(0, "Command %s killed by signal %d", cmd, WTERMSIG(st));
+	else
+		ploop_err(0, "Command %s died abnormally", cmd);
+
+	return SYSEXIT_SYS;
+}
+
 int send_process(const char *device, int ofd, const char *flush_cmd)
 {
 	struct delta idelta = { .fd = -1 };
@@ -112,6 +237,8 @@ int send_process(const char *device, int ofd, const char *flush_cmd)
 	char *send_from = NULL;
 	char *format = NULL;
 	void *iobuf = NULL;
+	int blocksize;
+	__u64 cluster;
 	__u64 pos;
 	__u64 iterpos;
 	__u64 trackpos;
@@ -120,11 +247,9 @@ int send_process(const char *device, int ofd, const char *flush_cmd)
 	int iter;
 	struct ploop_track_extent e;
 
-	if (posix_memalign(&iobuf, 4096, CLUSTER)) {
-		ploop_err(errno, "posix_memalign");
-		ret = SYSEXIT_MALLOC;
-		goto done;
-	}
+	// Do not print anything on stdout, since we use it to send delta
+	if (ofd == STDOUT_FILENO)
+		ploop_set_verbose_level(PLOOP_LOG_NOSTDOUT);
 
 	devfd = open(device, O_RDONLY);
 	if (devfd < 0) {
@@ -133,36 +258,51 @@ int send_process(const char *device, int ofd, const char *flush_cmd)
 		goto done;
 	}
 
-	if (ioctl(devfd, PLOOP_IOC_TRACK_INIT, &e)) {
+	ret = get_image_info(device, &send_from, &format, &blocksize);
+	if (ret)
+		goto done;
+	cluster = S2B(blocksize);
+
+	if (posix_memalign(&iobuf, 4096, cluster)) {
+		ploop_err(errno, "posix_memalign");
+		ret = SYSEXIT_MALLOC;
+		goto done;
+	}
+
+	iter = 0;
+	while (ioctl(devfd, PLOOP_IOC_TRACK_INIT, &e)) {
+		if (errno == EBUSY && iter++ == 0) {
+			/* Previous instance was killed?
+			 * Abort tracking and retry
+			 */
+			ioctl(devfd, PLOOP_IOC_TRACK_ABORT, 0);
+			continue;
+		}
 		ploop_err(errno, "PLOOP_IOC_TRACK_INIT");
 		ret = SYSEXIT_DEVIOC;
 		goto done;
 	}
 	tracker_on = 1;
 
-	if (get_image_info(device, &send_from, &format)) {
-		ret = SYSEXIT_OPEN;
-		goto done;
-	}
-
 	if (open_delta_simple(&idelta, send_from, O_RDONLY|O_DIRECT, OD_NOFLAGS)) {
 		ret = SYSEXIT_OPEN;
 		goto done;
 	}
 
-	trackend = e.end;
+	ploop_log(-1, "Sending %s", send_from);
 
+	trackend = e.end;
 	for (pos = 0; pos < trackend; ) {
 		int n;
 
-		trackpos = pos + CLUSTER;
+		trackpos = pos + cluster;
 		if (ioctl(devfd, PLOOP_IOC_TRACK_SETPOS, &trackpos)) {
-			ploop_err(errno, "PLOOP_IOC_TRACK_INIT");
+			ploop_err(errno, "PLOOP_IOC_TRACK_SETPOS");
 			ret = SYSEXIT_DEVIOC;
 			goto done;
 		}
 
-		n = idelta.fops->pread(idelta.fd, iobuf, CLUSTER, pos);
+		n = idelta.fops->pread(idelta.fd, iobuf, cluster, pos);
 		if (n < 0) {
 			ploop_err(errno, "pread");
 			ret = SYSEXIT_READ;
@@ -171,9 +311,9 @@ int send_process(const char *device, int ofd, const char *flush_cmd)
 		if (n == 0)
 			break;
 
-		if (send_buf(ofd, iobuf, n, pos)) {
+		ret = send_buf(ofd, iobuf, n, pos);
+		if (ret) {
 			ploop_err(errno, "write");
-			ret = SYSEXIT_WRITE;
 			goto done;
 		}
 
@@ -205,12 +345,12 @@ int send_process(const char *device, int ofd, const char *flush_cmd)
 				int n;
 				int copy = e.end - pos;
 
-				if (copy > CLUSTER)
-					copy = CLUSTER;
+				if (copy > cluster)
+					copy = cluster;
 				if (pos + copy > trackpos) {
 					trackpos = pos + copy;
 					if (ioctl(devfd, PLOOP_IOC_TRACK_SETPOS, &trackpos)) {
-						ploop_err(errno, "PLOOP_IOC_TRACK_INIT");
+						ploop_err(errno, "PLOOP_IOC_TRACK_SETPOS");
 						ret = SYSEXIT_DEVIOC;
 						goto done;
 					}
@@ -226,9 +366,9 @@ int send_process(const char *device, int ofd, const char *flush_cmd)
 					ret = SYSEXIT_READ;
 					goto done;
 				}
-				if (send_buf(ofd, iobuf, n, pos)) {
+				ret = send_buf(ofd, iobuf, n, pos);
+				if (ret) {
 					ploop_err(errno, "write2");
-					ret = SYSEXIT_WRITE;
 					goto done;
 				}
 				pos += n;
@@ -254,8 +394,9 @@ int send_process(const char *device, int ofd, const char *flush_cmd)
 	 * and suspend VE with subsequent fsyncing FS.
 	 */
 
-	if (flush_cmd)
-		system(flush_cmd);
+	ret = run_cmd(flush_cmd);
+	if (ret)
+		goto done;
 
 	if (ioctl(devfd, PLOOP_IOC_SYNC, 0)) {
 		ploop_err(errno, "PLOOP_IOC_SYNC");
@@ -286,12 +427,12 @@ int send_process(const char *device, int ofd, const char *flush_cmd)
 				int n;
 				int copy = e.end - pos;
 
-				if (copy > CLUSTER)
-					copy = CLUSTER;
+				if (copy > cluster)
+					copy = cluster;
 				if (pos + copy > trackpos) {
 					trackpos = pos + copy;
 					if (ioctl(devfd, PLOOP_IOC_TRACK_SETPOS, &trackpos)) {
-						ploop_err(errno, "PLOOP_IOC_TRACK_INIT");
+						ploop_err(errno, "PLOOP_IOC_TRACK_SETPOS");
 						ret = SYSEXIT_DEVIOC;
 						goto done;
 					}
@@ -307,9 +448,9 @@ int send_process(const char *device, int ofd, const char *flush_cmd)
 					ret = SYSEXIT_READ;
 					goto done;
 				}
-				if (send_buf(ofd, iobuf, n, pos)) {
+				ret = send_buf(ofd, iobuf, n, pos);
+				if (ret) {
 					ploop_err(errno, "write3");
-					ret = SYSEXIT_WRITE;
 					goto done;
 				}
 				pos += n;
@@ -335,21 +476,18 @@ int send_process(const char *device, int ofd, const char *flush_cmd)
 		int n;
 		struct ploop_pvd_header *vh = (void*)iobuf;
 
-		n = idelta.fops->pread(idelta.fd, iobuf, CLUSTER, 0);
-		if (n != CLUSTER) {
-			if (n < 0)
-				ploop_err(errno, "read header");
-			else
-				ploop_err(0, "short read header");
+		n = idelta.fops->pread(idelta.fd, iobuf, SECTOR_SIZE, 0);
+		if (n != SECTOR_SIZE) {
+			ploop_err(errno, "Error reading 1st sector of %s", send_from);
 			ret = SYSEXIT_READ;
 			goto done;
 		}
 
 		vh->m_DiskInUse = 0;
 
-		if (send_buf(ofd, iobuf, CLUSTER, 0)) {
+		ret = send_buf(ofd, vh, SECTOR_SIZE, 0);
+		if (ret) {
 			ploop_err(errno, "write3");
-			ret = SYSEXIT_WRITE;
 			goto done;
 		}
 	}
@@ -361,15 +499,9 @@ int send_process(const char *device, int ofd, const char *flush_cmd)
 	}
 	tracker_on = 0;
 
-	if (send_buf(ofd, iobuf, 0, 0)) {
+	ret = send_buf(ofd, iobuf, 0, 0);
+	if (ret) {
 		ploop_err(errno, "write4");
-		ret = SYSEXIT_WRITE;
-		goto done;
-	}
-
-	if (fsync(ofd)) {
-		ploop_err(errno, "fsync");
-		ret = SYSEXIT_WRITE;
 		goto done;
 	}
 
@@ -384,8 +516,6 @@ done:
 		close(devfd);
 	if (send_from)
 		free(send_from);
-	if (format)
-		free(format);
 	if (idelta.fd >= 0)
 		close_delta(&idelta);
 
