@@ -32,7 +32,6 @@
 #include <fcntl.h>
 #include <sys/vfs.h>
 #include <sys/syscall.h>
-#include <assert.h>
 #include <mntent.h>
 
 #include "ploop.h"
@@ -51,14 +50,14 @@ static int not_supported_for_vm(struct ploop_disk_images_data *di)
 	if (!di->runtime->vm_compat)
 		return 0;
 
-	ploop_err(0, "Operation is not supported for Vm");
+	ploop_err(0, "Operation is not supported for VM");
 	return SYSEXIT_PARAM;
 }
 
 /* set cancel flag
  * Note: this function also clear the flag
  */
-int is_operation_cancelled(void)
+static int is_operation_cancelled(void)
 {
 	struct ploop_cancel_handle *cancel_data;
 
@@ -92,6 +91,11 @@ int sys_fallocate(int fd, int mode, off_t offset, off_t len)
 	return syscall(__NR_fallocate, fd, mode, offset, len);
 }
 
+int sys_syncfs(int fd)
+{
+	return syscall(__NR_syncfs, fd);
+}
+
 int get_list_size(char **list)
 {
 	int i;
@@ -100,7 +104,7 @@ int get_list_size(char **list)
 	return i;
 }
 
-char **make_images_list(struct ploop_disk_images_data *di, char *guid, int reverse)
+static char **make_images_list(struct ploop_disk_images_data *di, char *guid, int reverse)
 {
 	int n;
 	char **images;
@@ -166,6 +170,20 @@ err:
 	return NULL;
 }
 
+static int get_snapshot_count(struct ploop_disk_images_data *di)
+{
+	int n;
+	char **images;
+
+	images = make_images_list(di, di->top_guid, 1);
+	if (images == NULL)
+		return -1;
+	n = get_list_size(images);
+	free_images_list(images);
+
+	return n;
+}
+
 void free_images_list(char **images)
 {
 	int i;
@@ -331,14 +349,10 @@ static int create_empty_preallocated_delta(const char *path, __u32 blocksize, of
 
 	rc = sys_fallocate(odelta.fd, 0, 0, S2B(vh.m_FirstBlockOffset + vh.m_SizeInSectors));
 	if (rc) {
-		if (errno == ENOTSUP) {
-			ploop_log(0, "Warning: fallocate is not supported, using truncate instead");
-			rc = ftruncate(odelta.fd, S2B(vh.m_FirstBlockOffset + vh.m_SizeInSectors));
-		}
-		if (rc) {
-			ploop_err(errno, "Failed to create %s", path);
-			goto out_close;
-		}
+		if (errno == ENOTSUP)
+			ploop_err(errno, "fallocate");
+		ploop_err(errno, "Failed to create %s", path);
+		goto out_close;
 	}
 
 	for (clu = 0; clu < SizeToFill / cluster; clu++) {
@@ -466,7 +480,7 @@ static int create_image(struct ploop_disk_images_data *di,
 
 	ret = SYSEXIT_NOMEM;
 	di->size = size_sec;
-	di->raw = mode == PLOOP_RAW_MODE;
+	di->mode = mode;
 
 	ret = SYSEXIT_CREAT;
 	if (mode == PLOOP_RAW_MODE)
@@ -521,18 +535,22 @@ static int create_balloon_file(struct ploop_disk_images_data *di,
 	mount_param.target = mnt;
 	ret = ploop_mount_fs(&mount_param);
 	if (ret)
-		return ret;
+		goto out;
 	snprintf(fname, sizeof(fname), "%s/"BALLOON_FNAME, mnt);
 
 	fd = open(fname, O_CREAT|O_RDONLY|O_TRUNC, 0600);
 	if (fd == -1) {
-		ploop_err(errno, "Can't open %s", mnt);
-		umount(mnt);
-		return SYSEXIT_OPEN;
+		ploop_err(errno, "Can't create balloon file %s", fname);
+		ret = SYSEXIT_CREAT;
+		goto out;
 	}
 	close(fd);
+	ret = 0;
+out:
+	umount(mnt);
+	rmdir(mnt);
 
-	return 0;
+	return ret;
 }
 
 static int ploop_init_image(struct ploop_disk_images_data *di, struct ploop_create_param *param)
@@ -576,7 +594,7 @@ err:
 	return ret;
 }
 
-int ploop_drop_image(struct ploop_disk_images_data *di)
+static int ploop_drop_image(struct ploop_disk_images_data *di)
 {
 	int i;
 	char fname[MAXPATHLEN];
@@ -627,9 +645,9 @@ int ploop_create_image(struct ploop_create_param *param)
 		ret = ploop_init_image(di, param);
 		if (ret)
 			ploop_drop_image(di);
-
-		ploop_free_diskdescriptor(di);
 	}
+	ploop_free_diskdescriptor(di);
+
 	return ret;
 }
 
@@ -664,7 +682,7 @@ int ploop_getdevice(int *minor)
 }
 
 /* Workaround for bug #PCLIN-30116 */
-int do_ioctl(int fd, int req)
+static int do_ioctl(int fd, int req)
 {
 	int i, ret;
 
@@ -677,7 +695,7 @@ int do_ioctl(int fd, int req)
 	return ret;
 }
 
-int do_umount(const char *mnt)
+static int do_umount(const char *mnt)
 {
 	int i, ret;
 
@@ -691,7 +709,7 @@ int do_umount(const char *mnt)
 	return ret;
 }
 
-int delete_deltas(int devfd, const char *devname)
+static int delete_deltas(int devfd, const char *devname)
 {
 	int top;
 
@@ -760,6 +778,11 @@ static int get_mount_dir(const char *dev, char *buf, int size)
 	return ret;
 }
 
+int ploop_get_mnt_by_dev(const char *dev, char *buf, int size)
+{
+	return get_mount_dir(dev, buf, size);
+}
+
 int ploop_fname_cmp(const char *p1, const char *p2)
 {
 	struct stat st1, st2;
@@ -1217,7 +1240,7 @@ static int mount_image(struct ploop_disk_images_data *di, struct ploop_mount_par
 	images = make_images_list(di, guid, 0);
 	if (images == NULL)
 		return SYSEXIT_NOMEM;
-	ret = ploop_mount(di, images, param, di->raw);
+	ret = ploop_mount(di, images, param, (di->mode == PLOOP_RAW_MODE));
 	free_images_list(images);
 
 	return ret;
@@ -1295,26 +1318,10 @@ int ploop_mount_snapshot(struct ploop_disk_images_data *di, struct ploop_mount_p
 	return mount_image(di, param, PLOOP_MOUNT_SNAPSHOT);
 }
 
-int ploop_umount(const char *device, struct ploop_disk_images_data *di)
+static int ploop_stop_device(const char *device)
 {
-	char mnt[MAXPATHLEN] = "";
 	int lfd, ret;
 
-	if (!device) {
-		ploop_err(0, "ploop_umount: device is not specified");
-		return -1;
-	}
-
-	if (get_mount_dir(device, mnt, sizeof(mnt)) == 0) {
-		if (di != NULL)
-			store_statfs_info(mnt, di->images[0]->file);
-		ploop_log(0, "Umounting fs at %s", mnt);
-		if (do_umount(mnt)) {
-			ploop_err(errno, "umount %s failed", mnt);
-			return SYSEXIT_UMOUNT;
-		}
-	}
-
 	ploop_log(0, "Unmounting device %s", device);
 	lfd = open(device, O_RDONLY);
 	if (lfd < 0) {
@@ -1325,9 +1332,41 @@ int ploop_umount(const char *device, struct ploop_disk_images_data *di)
 	ret = ploop_stop(lfd, device);
 	close(lfd);
 
+	return ret;
+}
+
+int ploop_umount(const char *device, struct ploop_disk_images_data *di)
+{
+	int ret;
+	char mnt[MAXPATHLEN] = "";
+	int lckfd;
+
+	if (!device) {
+		ploop_err(0, "ploop_umount: device is not specified");
+		return -1;
+	}
+
+	if (get_mount_dir(device, mnt, sizeof(mnt)) == 0) {
+		if (di != NULL)
+			store_statfs_info(mnt, di->images[0]->file);
+		ploop_log(0, "Unmounting file system at %s", mnt);
+		if (do_umount(mnt)) {
+			ploop_err(errno, "umount %s failed", mnt);
+			return SYSEXIT_UMOUNT;
+		}
+	}
+
+	lckfd = ploop_global_lock();
+	if (lckfd == -1)
+		return SYSEXIT_LOCK;
+
+	ret = ploop_stop_device(device);
+
 	if (ret == 0 && di != NULL)
 		unregister_ploop_dev(di->runtime->component_name, di->images[0]->file);
 
+	ploop_unlock(&lckfd);
+
 	return ret;
 }
 
@@ -1469,6 +1508,9 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 		ret = ploop_balloon_change_size(mount_param.device,
 				balloonfd, new_balloon_size);
 	} else if (param->size > dev_size) {
+		char conf[MAXPATHLEN];
+		char conf_tmp[MAXPATHLEN];
+
 		// GROW
 		if (balloon_size != 0) {
 			ret = ploop_balloon_change_size(mount_param.device,
@@ -1476,9 +1518,27 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 			if (ret)
 				goto err;
 		}
-		ret = ploop_grow_device(mount_param.device, di->blocksize, param->size);
+		// Update size in the DiskDescriptor.xml
+		di->size = param->size;
+		get_disk_descriptor_fname(di, conf, sizeof(conf));
+		snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
+		ret = ploop_store_diskdescriptor(conf_tmp, di);
 		if (ret)
 			goto err;
+
+		ret = ploop_grow_device(mount_param.device, di->blocksize, param->size);
+		if (ret) {
+			unlink(conf_tmp);
+			goto err;
+		}
+
+		if (rename(conf_tmp, conf)) {
+			ploop_err(errno, "Can't rename %s to %s",
+					conf_tmp, conf);
+			ret = SYSEXIT_RENAME;
+			goto err;
+		}
+
 		ret = resize_fs(mount_param.device);
 		if (ret)
 			goto err;
@@ -1584,7 +1644,6 @@ static int expanded2raw(struct ploop_disk_images_data *di)
 
 	if (fsync(odelta.fd))
 		ploop_err(errno, "fsync");
-	di->raw = 1;
 
 	get_disk_descriptor_fname(di, fname, sizeof(fname));
 	if (ploop_store_diskdescriptor(fname, di))
@@ -1647,11 +1706,9 @@ static int expanded2preallocated(struct ploop_disk_images_data *di)
 			ret = sys_fallocate(delta.fd, 0, data_off * cluster, cluster);
 			if (ret) {
 				if (errno == ENOTSUP)
-					ret = ftruncate(delta.fd, (data_off + 1) * cluster);
-				if (ret) {
-					ploop_err(errno, "Failed to expand %s", di->images[0]->file);
-					goto err;
-				}
+					ploop_err(errno, "fallocate");
+				ploop_err(errno, "Failed to expand %s", di->images[0]->file);
+				goto err;
 			}
 
 			if (PWRITE(&delta, &delta.l2[l2_slot], sizeof(__u32), idx_off))
@@ -1672,10 +1729,12 @@ err:
 
 int ploop_convert_image(struct ploop_disk_images_data *di, int mode, int flags)
 {
+	char conf_tmp[MAXPATHLEN];
+	char conf[MAXPATHLEN];
 	int ret = -1;
 
-	if (di->raw) {
-		ploop_err(0, "Only expanded image type is supported");
+	if (di->mode == PLOOP_RAW_MODE) {
+		ploop_err(0, "Converting raw image is not supported");
 		return SYSEXIT_PARAM;
 	}
 	if (di->nimages == 0) {
@@ -1685,11 +1744,32 @@ int ploop_convert_image(struct ploop_disk_images_data *di, int mode, int flags)
 	if (ploop_lock_di(di))
 		return SYSEXIT_LOCK;
 
+	di->mode = mode;
+	get_disk_descriptor_fname(di, conf, sizeof(conf));
+	snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
+	ret = ploop_store_diskdescriptor(conf_tmp, di);
+	if (ret)
+		goto err;
+
 	if (mode == PLOOP_EXPANDED_PREALLOCATED_MODE)
 		ret = expanded2preallocated(di);
 	else if (mode == PLOOP_RAW_MODE)
 		ret = expanded2raw(di);
+	/* else if (mode == PLOOP_EXPANDED)
+	 * do nothing because di->mode = mode and store DiskDescriptot.xml (see above) is enough;
+	 */
+	if (ret) {
+		unlink(conf_tmp);
+		goto err;
+	}
 
+	if (rename(conf_tmp, conf)) {
+		ploop_err(errno, "Can't rename %s %s",
+				conf_tmp, conf);
+		ret = SYSEXIT_RENAME;
+	}
+
+err:
 	ploop_unlock_di(di);
 
 	return ret;
@@ -1783,6 +1863,51 @@ err:
 	return ret;
 }
 
+static int get_image_size(struct ploop_disk_images_data *di, const char *guid, off_t *size)
+{
+	int ret;
+	struct delta delta;
+	const char *image;
+	int raw = 0;
+
+	image = find_image_by_guid(di, guid);
+	if (image == NULL) {
+		ploop_err(0, "Can't find image by top guid %s",
+				guid);
+		return SYSEXIT_PARAM;
+	}
+	if (di->mode == PLOOP_RAW_MODE) {
+		int i;
+
+		i = find_snapshot_by_guid(di, guid);
+		if (i == -1) {
+			ploop_err(0, "Can't find snapshot by guid %s",
+					guid);
+			return SYSEXIT_PARAM;
+		}
+		if (strcmp(di->snapshots[i]->parent_guid, NONE_UUID) == 0)
+			raw = 1;
+	}
+	if (raw) {
+		struct stat st;
+
+		if (stat(image, &st)) {
+			ploop_err(errno, "Failed to stat %s",
+					image);
+			return SYSEXIT_FSTAT;
+		}
+		*size = st.st_size / SECTOR_SIZE;
+	} else {
+		ret = open_delta(&delta, image, O_RDONLY, OD_OFFLINE);
+		if (ret)
+			return ret;
+		*size = delta.l2_size * delta.blocksize;
+		close_delta(&delta);
+	}
+
+	return 0;
+}
+
 int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapshot_param *param)
 {
 	int ret;
@@ -1794,6 +1919,9 @@ int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapsh
 	char conf[MAXPATHLEN];
 	char conf_tmp[MAXPATHLEN];
 	int online = 0;
+	int n;
+	off_t size;
+	char *top_guid = NULL;
 
 	ret = not_supported_for_vm(di);
 	if (ret)
@@ -1828,6 +1956,22 @@ int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapsh
 		else
 			strcpy(uuid, param->guid);
 	}
+	n = get_snapshot_count(di);
+	if (n == -1) {
+		ret = SYSEXIT_PARAM;
+		goto err_cleanup1;
+	} else if (n > 128-2) {
+		/* The number of images limited by 128
+		   so the snapshot limit 128 - base_image - one_reserverd
+		 */
+		ret = SYSEXIT_PARAM;
+		ploop_err(errno, "Unable to create a snapshot."
+			" The maximum number of snapshots (%d) has been reached.",
+			n-1);
+		goto err_cleanup1;
+	}
+
+	top_guid = strdup(di->top_guid);
 	snprintf(fname, sizeof(fname), "%s.%s",
 			di->images[0]->file, uuid1);
 	ret = ploop_di_add_image(di, fname, uuid, get_top_delta_guid(di));
@@ -1846,20 +1990,23 @@ int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapsh
 
 	if (ret != 0) {
 		// offline snapshot
-		fd = create_empty_delta(fname, di->blocksize, di->size);
-		if (fd != -1) {
-			ret = 0;
-			close(fd);
-		} else {
+		ret = get_image_size(di, top_guid, &size);
+		if (ret)
+			goto err_cleanup2;
+		fd = create_empty_delta(fname, di->blocksize, size);
+		if (fd < 0) {
 			ret = SYSEXIT_CREAT;
+			goto err_cleanup2;
 		}
+		close(fd);
 	} else {
 		// Always sync fs
 		online = 1;
 		ret = create_snapshot(dev, fname, di->blocksize, 1);
+		if (ret)
+			goto err_cleanup2;
+
 	}
-	if (ret)
-		goto err_cleanup2;
 
 	if (rename(conf_tmp, conf)) {
 		ploop_err(errno, "Can't rename %s %s",
@@ -1879,6 +2026,7 @@ err_cleanup2:
 				conf_tmp);
 err_cleanup1:
 	ploop_unlock_di(di);
+	free(top_guid);
 
 	return ret;
 }
@@ -1890,10 +2038,11 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 	char dev[64];
 	char uuid[61];
 	char uuid1[61];
-	char new_top_delta_fnanme[MAXPATHLEN];
+	char new_top_delta_fname[MAXPATHLEN];
 	char *old_top_delta_fname = NULL;
 	char conf[MAXPATHLEN];
 	char conf_tmp[MAXPATHLEN];
+	off_t size;
 
 	ret = not_supported_for_vm(di);
 	if (ret)
@@ -1914,6 +2063,11 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 				guid);
 		goto err_cleanup1;
 	}
+	// Read image size from image header
+	ret = get_image_size(di, guid, &size);
+	if (ret)
+		goto err_cleanup1;
+
 	ret = gen_uuid_pair(uuid, sizeof(uuid), uuid1, sizeof(uuid1));
 	if (ret) {
 		ploop_err(errno, "Can't generate uuid");
@@ -1936,9 +2090,9 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 		if (ret)
 			goto err_cleanup1;
 	}
-	snprintf(new_top_delta_fnanme, sizeof(new_top_delta_fnanme), "%s.%s",
+	snprintf(new_top_delta_fname, sizeof(new_top_delta_fname), "%s.%s",
 			di->images[0]->file, uuid1);
-	ret = ploop_di_add_image(di, new_top_delta_fnanme, uuid, guid);
+	ret = ploop_di_add_image(di, new_top_delta_fname, uuid, guid);
 	if (ret)
 		goto err_cleanup1;
 
@@ -1949,7 +2103,7 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 		goto err_cleanup1;
 
 	// offline snapshot
-	fd = create_empty_delta(new_top_delta_fnanme, di->blocksize, di->size);
+	fd = create_empty_delta(new_top_delta_fname, di->blocksize, size);
 	if (fd == -1) {
 		ret = SYSEXIT_CREAT;
 		goto err_cleanup2;
@@ -1971,7 +2125,7 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 
 	ploop_log(0, "ploop snapshot has been successfully switched");
 err_cleanup3:
-	if (ret && unlink(new_top_delta_fnanme))
+	if (ret && unlink(new_top_delta_fname))
 		ploop_err(errno, "Can't unlink %s",
 				conf_tmp);
 err_cleanup2:
