@@ -22,6 +22,16 @@
 #endif
 #endif /* ! __NR_fallocate */
 
+#ifndef __NR_syncfs
+#if defined __i386__
+#define __NR_syncfs	344
+#elif defined __x86_64__
+#define __NR_syncfs	306
+#else
+#error "No syncfs syscall known for this arch"
+#endif
+#endif /* ! __NR_syncfs */
+
 #define DISKDESCRIPTOR_XML      "DiskDescriptor.xml"
 /* Compatibility: Parallels use this UUID to mark top delta */
 #define BASE_UUID		"{5fbaabe3-6958-40ff-92a7-860e329aab41}"
@@ -160,39 +170,36 @@ enum {
 	PLOOP_MERGE_WITH_PARENT = 1,
 };
 
-extern int send_process(const char *device, int ofd, const char *flush_cmd);
+/* Mark lib functions used by ploop tools */
+#define PL_EXT __attribute__ ((visibility("default")))
+
+PL_EXT int send_process(const char *device, int ofd, const char *flush_cmd);
 
-extern int uuid_new(unsigned char * uuid);
 int gen_uuid_pair(char *uuid1, int len1, char *uuid2, int len2);
-extern const char *prl_uuid2str(unsigned char *in, char *out, int len);
 
-extern int find_delta_names(const char * device, int start_level, int end_level,
+int find_delta_names(const char * device, int start_level, int end_level,
 			    char **names, char ** format);
-extern int find_topdelta_name(const char *device, char **image);
-extern int ploop_get_attr(const char * device, char * attr, int * res);
-extern int ploop_get_delta_attr(const char * device, int level, char * attr, int * res);
-extern int ploop_get_delta_attr_str(const char * device, int level, char * attr, char *nbuf, int nbuf_len);
-extern int ploop_get_size(const char * device, off_t * res);
+PL_EXT int ploop_get_attr(const char * device, const char * attr, int * res);
+int ploop_get_delta_attr(const char * device, int level, char * attr, int * res);
+int ploop_get_size(const char * device, off_t * res);
 int dev_num2dev_start(const char *device, dev_t dev_num, __u32 *dev_start);
 int ploop_get_top_level(int devfd, const char *devname, int *top);
 
-extern int init_delta_array(struct delta_array *);
+int init_delta_array(struct delta_array *);
 void deinit_delta_array(struct delta_array * p);
-extern int extend_delta_array(struct delta_array * p, char * path, int rw, int od_flags);
-extern void close_delta(struct delta *delta);
-extern int open_delta(struct delta * delta, const char * path, int rw, int od_flags);
-extern int open_delta_simple(struct delta * delta, const char * path, int rw, int od_flags);
-extern int dirty_delta(struct delta * delta);
-extern int clear_delta(struct delta * delta);
-extern int parse_size(char * opt, off_t * sz);
-extern int read_size_from_image(const char *img_name, int raw, off_t * res);
-extern int grow_delta(struct delta *odelta, off_t bdsize, void *buf,
+int extend_delta_array(struct delta_array * p, char * path, int rw, int od_flags);
+PL_EXT void close_delta(struct delta *delta);
+PL_EXT int open_delta(struct delta * delta, const char * path, int rw, int od_flags);
+PL_EXT int open_delta_simple(struct delta * delta, const char * path, int rw, int od_flags);
+PL_EXT int dirty_delta(struct delta * delta);
+PL_EXT int clear_delta(struct delta * delta);
+PL_EXT int read_size_from_image(const char *img_name, int raw, off_t * res);
+PL_EXT int grow_delta(struct delta *odelta, off_t bdsize, void *buf,
 		       struct grow_maps *gm);
-extern int grow_raw_delta(const char *image, off_t append_size);
-int ploop_grow_device(const char *device, __u32 blocksize, off_t new_size);
+PL_EXT int grow_raw_delta(const char *image, off_t append_size);
+PL_EXT int ploop_grow_device(const char *device, __u32 blocksize, off_t new_size);
 
 struct pfiemap *fiemap_alloc(int n);
-int fiemap_add_extent(struct pfiemap **pfiemap_pp, __u64 pos, __u64 len);
 int fiemap_get(int fd, __u64 off, __u64 start, off_t size, struct pfiemap **pfiemap_pp);
 void fiemap_adjust(struct pfiemap *pfiemap, __u32 blocksize);
 int fiemap_build_rmap(struct pfiemap *pfiemap, __u32 *rmap, __u32 rlen, struct delta *delta);
@@ -207,8 +214,6 @@ int freemap2freeblks(struct freemap *freemap,
 int freeblks2freemap(struct ploop_freeblks_ctl *freeblks,
 		       struct freemap **freemap_pp, __u32 *total);
 
-int range_build_rmap(__u32 iblk_start, __u32 iblk_end,
-		       __u32 *rmap, __u32 rlen, struct delta *delta, __u32 *out);
 int range_build(__u32 a_h, __u32 n_free_blocks,
 		__u32 *rmap, __u32 rlen,
 		struct delta     *delta,
@@ -216,18 +221,11 @@ int range_build(__u32 a_h, __u32 n_free_blocks,
 		struct freemap  **rangemap_pp,
 		struct relocmap **relocmap_pp);
 
-void range_fix_gaps(struct freemap *freemap, __u32 iblk_start, __u32 iblk_end,
-		    __u32 n_to_fix, __u32 *rmap);
-int range_split(struct freemap *rangemap, struct freemap *freemap,
-		 struct relocmap **relocmap_pp);
-
 struct relocmap *relocmap_alloc(int n);
-int relocmap_add_extent(struct relocmap **relocmap_pp,
-			 __u32 clu, __u32 iblk, __u32 len, __u32 free);
 struct ploop_relocblks_ctl;
 int relocmap2relocblks(struct relocmap *relocmap, int lvl, __u32 a_h, __u32 n_scanned,
 			struct ploop_relocblks_ctl **relocblks_pp);
-int ploop_fsck(char *img, int flags, int ro, int verbose, __u32 *blocksize_p);
+PL_EXT int ploop_fsck(char *img, int flags, int ro, int verbose, __u32 *blocksize_p);
 /* Logging */
 void ploop_log(int level, const char *format, ...)
 	__attribute__ ((__format__ (__printf__, 2, 3)));
@@ -244,22 +242,18 @@ void __ploop_err(int err_no, const char *format, ...)
 #endif
 
 char *make_sysfs_dev_name(int minor, char *buf, int len);
-int ploop_mount(struct ploop_disk_images_data *di, char **images,
+PL_EXT int ploop_mount(struct ploop_disk_images_data *di, char **images,
 		struct ploop_mount_param *param, int raw);
-int create_snapshot(const char *device, const char *delta, __u32 blocksize, int syncfs);
+PL_EXT int create_snapshot(const char *device, const char *delta, __u32 blocksize, int syncfs);
 int get_list_size(char **list);
-char **make_images_list(struct ploop_disk_images_data *di, char *guid, int reverse);
 void free_images_list(char **images);
 int PWRITE(struct delta * delta, void * buf, unsigned int size, off_t off);
 int PREAD(struct delta * delta, void *buf, unsigned int size, off_t off);
-int ploop_get_dev_by_mnt(const char *path, char *buf, int size);
-int ploop_remove_images(struct ploop_disk_images_data *di, char **images, char ***images_out);
 void get_disk_descriptor_fname(struct ploop_disk_images_data *di, char *buf, int size);
 void get_disk_descriptor_lock_fname(struct ploop_disk_images_data *di, char *out, int size);
 int ploop_find_dev_by_uuid(struct ploop_disk_images_data *di, int check_state, char *out, int len);
 int sys_fallocate(int fd, int mode, off_t offset, off_t len);
-
-int delete_deltas(int devfd, const char *devname);
+int sys_syncfs(int fd);
 
 // manage struct ploop_disk_images_data
 int ploop_di_add_image(struct ploop_disk_images_data *di, const char *fname,
@@ -277,15 +271,15 @@ int register_ploop_dev(const char *module, const char *image, const char *dev);
 void unregister_ploop_dev(const char *module, const char *image);
 
 //balloon
-char *mntn2str(int mntn_type);
-int get_balloon(const char *mount_point, struct stat *st, int *outfd);
-int ploop_balloon_change_size(const char *device, int balloonfd, off_t new_size);
-int ploop_balloon_change_size_local(const char *device, int balloonfd, off_t new_size);
-int ploop_balloon_get_state(const char *device, __u32 *state);
-int ploop_balloon_clear_state(const char *device);
-int ploop_balloon_complete(const char *device);
-int ploop_balloon_check_and_repair(const char *device, char *mount_point, int repair);
-int ploop_discard(const char *device, const char *mount_point);
+PL_EXT char *mntn2str(int mntn_type);
+PL_EXT int get_balloon(const char *mount_point, struct stat *st, int *outfd);
+PL_EXT int ploop_balloon_change_size(const char *device, int balloonfd, off_t new_size);
+PL_EXT int ploop_balloon_get_state(const char *device, __u32 *state);
+PL_EXT int ploop_balloon_clear_state(const char *device);
+PL_EXT int ploop_balloon_complete(const char *device);
+PL_EXT int ploop_balloon_check_and_repair(const char *device, char *mount_point, int repair);
+PL_EXT int ploop_discard(const char *device, const char *mount_point,
+		__u64 minblock, __u64 to_free);
 
 /* lock */
 int ploop_lock_di(struct ploop_disk_images_data *di);
@@ -302,7 +296,6 @@ int create_gpt_partition(const char *dev, off_t size, __u32 blocksize);
 int resize_gpt_partition(const char *devname);
 
 // misc
-int do_ioctl(int fd, int req);
 void get_basedir(const char *fname, char *out, int len);
 __u32 crc32(const unsigned char *buf, unsigned long len);
 int ploop_is_on_nfs(const char *path);
@@ -317,10 +310,8 @@ int read_line(const char *path, char *nbuf, int len);
 int is_valid_blocksize(__u32 blocksize);
 
 // merge
-int get_delta_info(const char *device, int merge_top_only, struct merge_info *info);
-int merge_image(const char *device, int start_level, int end_level, int raw, int merge_top,
-		char **images);
-int merge_image_local(const char *device, int start_level, int end_level, int raw, int merge_top,
+PL_EXT int get_delta_info(const char *device, struct merge_info *info);
+PL_EXT int merge_image(const char *device, int start_level, int end_level, int raw, int merge_top,
 		char **images);
 int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *guid, int merge_mode);
 
