@@ -28,17 +28,46 @@
 #include <getopt.h>
 #include <linux/types.h>
 #include <string.h>
+#include <limits.h>
 
 #include "ploop.h"
+#include "common.h"
 
-char  *mount_point;
-int    force;	   /* do not flock hidden balloon */
+static char  *device;	   /* ploop device name (e.g. /dev/ploop0) */
+static char  *mount_point;
+static int    force;	   /* do not flock hidden balloon */
 
-char  *device;	   /* ploop device name (e.g. /dev/ploop0) */
+static int fill_opts(void)
+{
+	static char _device[PATH_MAX];
+	static char _mount_point[PATH_MAX];
 
-struct delta delta;
+	if (device && mount_point)
+		return 0;
 
-void usage_summary(void)
+	if (!device && mount_point) {
+		if (ploop_get_dev_by_mnt(mount_point, _device, sizeof(_device))) {
+			fprintf(stderr, "Unable to find ploop device by %s\n", mount_point);
+			return -1;
+		}
+		device = _device;
+		return 0;
+	}
+
+	if (!mount_point && device) {
+		if (ploop_get_mnt_by_dev(device, _mount_point, sizeof(_mount_point))) {
+			fprintf(stderr, "Unable to find mount point for %s\n", device);
+			return -1;
+		}
+		mount_point = _mount_point;
+		return 0;
+	}
+
+	fprintf(stderr, "At least one of -d or -m is required\n");
+	return 1;
+}
+
+static void usage_summary(void)
 {
 	fprintf(stderr, "Usage: ploop-balloon { show | status | clear | change | complete | check | repair | discard } ...\n"
 			"Use \"ploop-balloon cmd\" to get more info about cmd\n"
@@ -47,7 +76,8 @@ void usage_summary(void)
 
 static void usage_show(void)
 {
-	fprintf(stderr, "Usage: ploop-balloon show [-f] -m MOUNT_POINT\n"
+	fprintf(stderr, "Usage: ploop-balloon show [-f] {-d DEVICE | -m MOUNT_POINT}\n"
+			"	DEVICE	    := ploop device, e.g. /dev/ploop0\n"
 			"	MOUNT_POINT := path where fs living on ploop device mounted to\n"
 			"	-f	    - do not flock hidden balloon\n"
 			"Action: show current ploop balloon size\n"
@@ -59,10 +89,13 @@ static int pb_show(int argc, char **argv)
 	int i, ret;
 	struct stat st;
 
-	while ((i = getopt(argc, argv, "fm:")) != EOF) {
+	while ((i = getopt(argc, argv, "fd:m:")) != EOF) {
 		switch (i) {
 		case 'f':
 			force = 1;
+		case 'd':
+			device = optarg;
+			break;
 		case 'm':
 			mount_point = optarg;
 			break;
@@ -75,7 +108,7 @@ static int pb_show(int argc, char **argv)
 	argc -= optind;
 	argv += optind;
 
-	if (argc != 0 || mount_point == NULL) {
+	if (argc != 0 || fill_opts()) {
 		usage_show();
 		return -1;
 	}
@@ -83,15 +116,14 @@ static int pb_show(int argc, char **argv)
 	ret = get_balloon(mount_point, &st, NULL);
 	if (ret)
 		return ret;
-	fprintf(stdout, "Current size of hidden balloon is %llu bytes",
+	fprintf(stdout, "Current size of hidden balloon is %llu bytes\n",
 		(unsigned long long) st.st_size);
 	return 0;
 }
 
 static void usage_status(void)
 {
-	fprintf(stderr, "Usage: ploop-balloon status -d DEVICE -m MOUNT_POINT\n"
-			"	ploop-balloon status -f -d DEVICE\n"
+	fprintf(stderr, "Usage: ploop-balloon status [-f] {-d DEVICE | -m MOUNT_POINT}\n"
 			"	DEVICE	    := ploop device, e.g. /dev/ploop0\n"
 			"	MOUNT_POINT := path where fs living on ploop device mounted to\n"
 			"	-f	    - do not flock hidden balloon\n"
@@ -123,9 +155,7 @@ static int pb_status(int argc, char **argv)
 	argc -= optind;
 	argv += optind;
 
-	if (argc != 0 || device == NULL ||
-	    (mount_point == NULL && !force) ||
-	    (mount_point != NULL && force)) {
+	if (argc != 0 || fill_opts()) {
 		usage_status();
 		return -1;
 	}
@@ -134,14 +164,14 @@ static int pb_status(int argc, char **argv)
 	if (ret)
 		return ret;
 
-	fprintf(stdout, "Current state of in-kernel maintenance: %s",
+	fprintf(stdout, "Current state of in-kernel maintenance: %s\n",
 	       mntn2str(state));
 	return 0;
 }
 
 static void usage_clear(void)
 {
-	fprintf(stderr, "Usage: ploop-balloon clear -d DEVICE -m MOUNT_POINT\n"
+	fprintf(stderr, "Usage: ploop-balloon clear {-d DEVICE | -m MOUNT_POINT}\n"
 			"	DEVICE	    := ploop device, e.g. /dev/ploop0\n"
 			"	MOUNT_POINT := path where fs living on ploop device mounted to\n"
 			"Action: clear stale in-kernel \"BALLOON\" state of maintenance\n"
@@ -170,7 +200,7 @@ static int pb_clear(int argc, char **argv)
 	argc -= optind;
 	argv += optind;
 
-	if (argc != 0 || device == NULL || mount_point == NULL) {
+	if (argc != 0 || fill_opts()) {
 		usage_clear();
 		return -1;
 	}
@@ -182,13 +212,13 @@ static int pb_clear(int argc, char **argv)
 	ret = ploop_balloon_clear_state(device);
 	if (ret)
 		return ret;
-	fprintf(stdout, "Current state of in-kernel maintenance is OFF now");
+	fprintf(stdout, "Current state of in-kernel maintenance is OFF now\n");
 	return 0;
 }
 
 static void usage_change(void)
 {
-	fprintf(stderr, "Usage: ploop-balloon change -s SIZE -d DEVICE -m MOUNT_POINT\n"
+	fprintf(stderr, "Usage: ploop-balloon change -s SIZE {-d DEVICE | -m MOUNT_POINT}\n"
 			"	SIZE	    := NUMBER[kmg] (new size of balloon)\n"
 			"	DEVICE	    := ploop device, e.g. /dev/ploop0\n"
 			"	MOUNT_POINT := path where fs living on ploop device mounted to\n"
@@ -228,8 +258,7 @@ static int pb_change(int argc, char **argv)
 	argc -= optind;
 	argv += optind;
 
-	if (argc != 0 || !new_size_set ||
-	    device == NULL || mount_point == NULL) {
+	if (argc != 0 || !new_size_set || fill_opts()) {
 		usage_change();
 		return -1;
 	}
@@ -242,7 +271,7 @@ static int pb_change(int argc, char **argv)
 
 static void usage_complete(void)
 {
-	fprintf(stderr, "Usage: ploop-balloon complete -d DEVICE -m MOUNT_POINT\n"
+	fprintf(stderr, "Usage: ploop-balloon complete {-d DEVICE | -m MOUNT_POINT}\n"
 			"	DEVICE	    := ploop device, e.g. /dev/ploop0\n"
 			"	MOUNT_POINT := path where fs living on ploop device mounted to\n"
 			"Action: complete previously interrupted balloon operation\n"
@@ -272,7 +301,7 @@ static int pb_complete(int argc, char **argv)
 	argc -= optind;
 	argv += optind;
 
-	if (argc != 0 || device == NULL || mount_point == NULL) {
+	if (argc != 0 || fill_opts()) {
 		usage_complete();
 		return -1;
 	}
@@ -286,7 +315,7 @@ static int pb_complete(int argc, char **argv)
 
 static void usage_check(void)
 {
-	fprintf(stderr, "Usage: ploop-balloon check -d DEVICE -m MOUNT_POINT\n"
+	fprintf(stderr, "Usage: ploop-balloon check {-d DEVICE | -m MOUNT_POINT}\n"
 			"	DEVICE	    := ploop device, e.g. /dev/ploop0\n"
 			"	MOUNT_POINT := path where fs living on ploop device mounted to\n"
 			"Action: check whether hidden balloon has free blocks\n"
@@ -295,7 +324,7 @@ static void usage_check(void)
 
 static void usage_repair(void)
 {
-	fprintf(stderr, "Usage: ploop-balloon repair -d DEVICE -m MOUNT_POINT\n"
+	fprintf(stderr, "Usage: ploop-balloon repair {-d DEVICE | -m MOUNT_POINT}\n"
 			"	DEVICE	    := ploop device, e.g. /dev/ploop0\n"
 			"	MOUNT_POINT := path where fs living on ploop device mounted to\n"
 			"Action: repair hidden balloon (i.e. relocate all free blocks\n"
@@ -329,9 +358,7 @@ static int pb_check_and_repair(int argc, char **argv, int repair)
 	argc -= optind;
 	argv += optind;
 
-	if (argc != 0 || device == NULL ||
-	    (mount_point == NULL && !force) ||
-	    (mount_point != NULL && force)) {
+	if (argc != 0 || fill_opts()) {
 		if (repair)
 			usage_repair();
 		else
@@ -345,9 +372,12 @@ static int pb_check_and_repair(int argc, char **argv, int repair)
 
 static void usage_discard(void)
 {
-	fprintf(stderr, "Usage: ploop-balloon discard -d DEVICE -m MOUNT_POINT\n"
-			"	DEVICE	    := ploop device, e.g. /dev/ploop0\n"
-			"	MOUNT_POINT := path where fs living on ploop device mounted to\n"
+	fprintf(stderr, "Usage: ploop-balloon discard {-d DEVICE | -m MOUNT_POINT}\n"
+			"                     --to-free SIZE --min-block MIN_SIZE\n"
+			"       DEVICE      := ploop device, e.g. /dev/ploop0\n"
+			"       MOUNT_POINT := path where fs living on ploop device mounted to\n"
+			"       SIZE        := NUMBER[kmg] (maximum space to free)\n"
+			"       MIN_SIZE    := NUMBER[kmg] (minimum size of a linear slice to be freed)\n"
 			"Action: discard unused blocks from the image.\n"
 		);
 }
@@ -355,8 +385,15 @@ static void usage_discard(void)
 static int pb_discard(int argc, char **argv)
 {
 	int i;
+	off_t val;
+	__u64 to_free = ~0ULL, minblock_b = 0;
+	static struct option long_opts[] = {
+		{ "to-free", required_argument, 0, 666 },
+		{ "min-block", required_argument, 0, 667 },
+		{},
+	};
 
-	while ((i = getopt(argc, argv, "d:m:")) != EOF) {
+	while ((i = getopt_long(argc, argv, "d:m:", long_opts, NULL)) != EOF) {
 		switch (i) {
 		case 'd':
 			device = optarg;
@@ -364,7 +401,24 @@ static int pb_discard(int argc, char **argv)
 		case 'm':
 			mount_point = optarg;
 			break;
+		case 666:
+			if (parse_size(optarg, &val)) {
+				fprintf(stderr, "Invalid value for --to-free: %s\n", optarg);
+				usage_discard();
+				return SYSEXIT_PARAM;
+			}
+			to_free = S2B(val);
+			break;
+		case 667:
+			if (parse_size(optarg, &val)) {
+				fprintf(stderr, "Invalid value for --min-block: %s\n", optarg);
+				usage_discard();
+				return SYSEXIT_PARAM;
+			}
+			minblock_b = S2B(val);
+			break;
 		default:
+			usage_discard();
 			return -1;
 		}
 	}
@@ -372,12 +426,12 @@ static int pb_discard(int argc, char **argv)
 	argc -= optind;
 	argv += optind;
 
-	if (argc != 0 || device == NULL || mount_point == NULL) {
+	if (argc != 0 || fill_opts()) {
 		usage_discard();
 		return -1;
 	}
 
-	return ploop_discard(device, mount_point);
+	return ploop_discard(device, mount_point, minblock_b, to_free);
 }
 
 int main(int argc, char **argv)
