@@ -29,13 +29,14 @@
 #include <signal.h>
 
 #include "ploop.h"
+#include "common.h"
 
 static struct ploop_cancel_handle *_s_cancel_handle;
 
-void usage_summary(void)
+static void usage_summary(void)
 {
 	fprintf(stderr, "Usage: ploop init -s SIZE [-f FORMAT] NEW_DELTA\n"
-			"       ploop mount [-rP] [-f raw] -d DEVICE BASE_DELTA [ ... TOP_DELTA ]\n"
+			"       ploop mount [-rP] [-f raw] [-d DEVICE] BASE_DELTA [ ... TOP_DELTA ]\n"
 			"       ploop umount -d DEVICE\n"
 			"       ploop { delete | rm } -d DEVICE -l LEVEL\n"
 			"       ploop merge -d DEVICE [-l LEVEL[..TOP_LEVEL]]\n"
@@ -61,19 +62,9 @@ static void usage_init(void)
 			"       FSTYPE := make file system\n");
 }
 
-static int is_xml_fname(const char *fname)
-{
-	char *p;
-
-	p = strrchr(fname, '.');
-	if (p != NULL && !strcmp(p, ".xml"))
-		return 1;
-	return 0;
-}
-
 static int plooptool_init(int argc, char **argv)
 {
-	int i, ret;
+	int i, f, ret;
 	off_t size_sec = 0;
 	struct ploop_create_param param = {};
 
@@ -90,24 +81,19 @@ static int plooptool_init(int argc, char **argv)
 			  char * endptr;
 
 			  param.blocksize = strtoul(optarg, &endptr, 0);
-			  if (optarg == endptr) {
+			  if (*endptr != '\0') {
 				  usage_init();
 				  return -1;
 			  }
 			  break;
 		}
 		case 'f':
-			if (strcmp(optarg, "raw") == 0)
-				param.mode = PLOOP_RAW_MODE;
-			else if ((strcmp(optarg, "ploop1") == 0) ||
-				 (strcmp(optarg, "expanded") == 0))
-				param.mode = PLOOP_EXPANDED_MODE;
-			else if (strcmp(optarg, "preallocated") == 0)
-				param.mode = PLOOP_EXPANDED_PREALLOCATED_MODE;
-			else {
+			f = parse_format_opt(optarg);
+			if (f < 0) {
 				usage_init();
 				return -1;
 			}
+			param.mode = f;
 			break;
 		case 't':
 			if (!strcmp(optarg, "ext4") || !strcmp(optarg, "ext3")) {
@@ -142,28 +128,30 @@ static int plooptool_init(int argc, char **argv)
 
 static void usage_mount(void)
 {
-	fprintf(stderr, "Usage: ploop mount [-rP] [-f FORMAT] [-b BLOCKSIZE] [-d DEVICE]\n"
+	fprintf(stderr, "Usage: ploop mount [-r] [-f FORMAT] [-b BLOCKSIZE] [-d DEVICE]\n"
+			"             [-m MOUNT_POINT] [-t FSTYPE] [-o MOUNT_OPTS]\n"
 			"             BASE_DELTA [ ... TOP_DELTA ]\n"
-			"       ploop mount [-rP] [-m DIR] [-u UUID] DiskDescriptor.xml\n"
+			"       ploop mount [-r] [-m MOUNT_POINT] [-u UUID] DiskDescriptor.xml\n"
 			"       FORMAT := { raw | ploop1 }\n"
 			"       BLOCKSIZE := block size (for raw image format)\n"
 			"       DEVICE := ploop device, e.g. /dev/ploop0\n"
+			"       MOUNT_POINT := directory to mount in-image filesystem to\n"
+			"       FSTYPE := in-image filesystem type (ext4 by default)\n"
+			"       MOUNT_OPTS := additional mount options, comma-separated\n"
 			"       *DELTA := path to image file\n"
 			"       -r     - mount images read-only\n"
-			"       -P     - rescan partition table upon mount\n"
-			"       -t     - file system type\n"
-			"       -m     - target mount point\n"
 		);
 }
 
 static int plooptool_mount(int argc, char **argv)
 {
-	int i, ret;
+	int i, f, ret;
 	int raw = 0;
 	int base = 0;
 	struct ploop_mount_param mountopts = {};
+	const char *component_name = NULL;
 
-	while ((i = getopt(argc, argv, "rf:Pd:m:t:u:o:b:")) != EOF) {
+	while ((i = getopt(argc, argv, "rf:d:m:t:u:o:b:c:")) != EOF) {
 		switch (i) {
 		case 'd':
 			strncpy(mountopts.device, optarg, sizeof(mountopts.device)-1);
@@ -172,15 +160,12 @@ static int plooptool_mount(int argc, char **argv)
 			mountopts.ro = 1;
 			break;
 		case 'f':
-			if (strcmp(optarg, "raw") == 0)
-				raw = 1;
-			else if (strcmp(optarg, "ploop1") != 0) {
+			f = parse_format_opt(optarg);
+			if (f < 0) {
 				usage_mount();
 				return -1;
 			}
-			break;
-		case 'P':
-			mountopts.partitioned = 1;
+			raw = (f == PLOOP_RAW_MODE);
 			break;
 		case 'm':
 			mountopts.target = strdup(optarg);
@@ -201,11 +186,14 @@ static int plooptool_mount(int argc, char **argv)
 			  char * endptr;
 
 			  mountopts.blocksize = strtoul(optarg, &endptr, 0);
-			  if (optarg == endptr) {
+			  if (*endptr != '\0') {
 				  usage_mount();
 				  return -1;
 			  }
 			  break;
+		case 'c':
+			component_name = optarg;
+			break;
 		}
 		default:
 			usage_mount();
@@ -228,11 +216,13 @@ static int plooptool_mount(int argc, char **argv)
 		ret = ploop_read_diskdescriptor(argv[0], di);
 		if (ret)
 			goto err;
+		if (component_name != NULL)
+			ploop_set_component_name(di, component_name);
 		if (base) {
 			mountopts.guid = ploop_get_base_delta_uuid(di);
 			if (mountopts.guid == NULL) {
 				ret = 1;
-				fprintf(stderr, "Unable to find base delta uuid");
+				fprintf(stderr, "Unable to find base delta uuid\n");
 				goto err;
 			}
 		}
@@ -249,9 +239,9 @@ err:
 
 static void usage_start(void)
 {
-	fprintf(stderr, "Usage: ploop start [-P] -d DEVICE\n"
+	fprintf(stderr, "Usage: ploop start -d DEVICE\n"
 			"       DEVICE := ploop device, e.g. /dev/ploop0\n"
-			"       -P     - rescan partition table\n");
+			);
 }
 
 static int plooptool_start(int argc, char **argv)
@@ -259,18 +249,14 @@ static int plooptool_start(int argc, char **argv)
 	int i;
 	int lfd;
 	struct {
-		int partitioned;
 		char * device;
 	} startopts = { };
 
-	while ((i = getopt(argc, argv, "d:P")) != EOF) {
+	while ((i = getopt(argc, argv, "d:")) != EOF) {
 		switch (i) {
 		case 'd':
 			startopts.device = optarg;
 			break;
-		case 'P':
-			startopts.partitioned = 1;
-			break;
 		default:
 			usage_start();
 			return -1;
@@ -296,9 +282,8 @@ static int plooptool_start(int argc, char **argv)
 		return SYSEXIT_DEVIOC;
 	}
 
-	if (startopts.partitioned && ioctl(lfd, BLKRRPART, 0) < 0) {
+	if (ioctl(lfd, BLKRRPART, 0) < 0) {
 		perror("BLKRRPART");
-		return SYSEXIT_BLKDEV;
 	}
 
 	return 0;
@@ -344,7 +329,7 @@ static int plooptool_stop(int argc, char **argv)
 	}
 
 	if (ioctl(lfd, PLOOP_IOC_STOP, 0) < 0) {
-		perror("PLOOP_IOC_START");
+		perror("PLOOP_IOC_STOP");
 		return SYSEXIT_DEVIOC;
 	}
 
@@ -391,7 +376,7 @@ static int plooptool_clear(int argc, char **argv)
 	}
 
 	if (ioctl(lfd, PLOOP_IOC_CLEAR, 0) < 0) {
-		perror("PLOOP_IOC_START");
+		perror("PLOOP_IOC_CLEAR");
 		return SYSEXIT_DEVIOC;
 	}
 
@@ -401,7 +386,12 @@ static int plooptool_clear(int argc, char **argv)
 static void usage_umount(void)
 {
 	fprintf(stderr, "Usage: ploop umount -d DEVICE\n"
-			"       DEVICE := ploop device, e.g. /dev/ploop0\n");
+			"       ploop umount -m DIR\n"
+			"       ploop umount DiskDescriptor.xml\n"
+			"       ploop umount DELTA\n"
+			"       DEVICE := ploop device, e.g. /dev/ploop0\n"
+			"       DIR := mount point\n"
+			"       DELTA := path to (mounted) image file\n");
 }
 
 static int plooptool_umount(int argc, char **argv)
@@ -439,7 +429,8 @@ static int plooptool_umount(int argc, char **argv)
 		ret = ploop_umount(umountopts.device, NULL);
 	}else if (mnt != NULL) {
 		if (ploop_get_dev_by_mnt(mnt, device, sizeof(device))) {
-			ploop_err(0, "Unable to find ploop device by %s", mnt);
+			fprintf(stderr, "Unable to find ploop device by %s\n",
+					mnt);
 			return -1;
 		}
 		ret = ploop_umount(device, NULL);
@@ -452,7 +443,7 @@ static int plooptool_umount(int argc, char **argv)
 		ploop_free_diskdescriptor(di);
 	} else {
 		if (ploop_find_dev_by_delta(argv[0], device, sizeof(device)) != 0) {
-			ploop_err(0, "Image %s is not mounted", argv[0]);
+			fprintf(stderr, "Image %s is not mounted\n", argv[0]);
 			return -1;
 		}
 		ret = ploop_umount(device, NULL);
@@ -531,7 +522,7 @@ static int plooptool_snapshot(int argc, char **argv)
 	int syncfs = 0;
 	struct ploop_snapshot_param param = {};
 
-	while ((i = getopt(argc, argv, "b:Fd:u:")) != EOF) {
+	while ((i = getopt(argc, argv, "Fd:u:")) != EOF) {
 		switch (i) {
 		case 'd':
 			device = optarg;
@@ -572,7 +563,7 @@ static int plooptool_snapshot(int argc, char **argv)
 			return -1;
 		}
 		if (ploop_get_attr(device, "block_size", (int*) &blocksize))
-			return 1;
+			return -1;
 		ret = create_snapshot(device, argv[0], blocksize, syncfs);
 	}
 
@@ -665,13 +656,13 @@ static int plooptool_snapshot_delete(int argc, char **argv)
 	return ret;
 }
 
-void usage_snapshot_merge(void)
+static void usage_snapshot_merge(void)
 {
-	fprintf(stderr, "Usage: ploop snapshot-merge [-u <uuid>] DiskDescriptor.xml]\n"
+	fprintf(stderr, "Usage: ploop snapshot-merge [-u <uuid> | -A] DiskDescriptor.xml\n"
 			"       -u <uuid>     snapshot uuid (top delta if not specified)\n");
 }
 
-int plooptool_snapshot_merge(int argc, char ** argv)
+static int plooptool_snapshot_merge(int argc, char ** argv)
 {
 	int i, ret;
 	struct ploop_merge_param param = {};
@@ -694,11 +685,12 @@ int plooptool_snapshot_merge(int argc, char ** argv)
 	argv += optind;
 
 	if (param.guid != NULL && param.merge_all != 0) {
+		fprintf(stderr, "Options -u and -A can't be used together\n");
 		usage_snapshot_merge();
-		return 1;
+		return -1;
 	}
 
-	if (argc == 1 && strstr(argv[0], DISKDESCRIPTOR_XML)) {
+	if (argc == 1 && is_xml_fname(argv[0])) {
 		struct ploop_disk_images_data *di;
 
 
@@ -709,7 +701,7 @@ int plooptool_snapshot_merge(int argc, char ** argv)
 		ploop_free_diskdescriptor(di);
 	} else {
 		usage_snapshot_merge();
-		return 1;
+		return -1;
 	}
 
 	return ret;
@@ -792,7 +784,9 @@ static int plooptool_resize(int argc, char **argv)
 
 static void usage_convert(void)
 {
-	fprintf(stderr, "Usage: ploop convert -t <raw|preallocated>\n");
+	fprintf(stderr, "Usage: ploop convert -f FORMAT\n"
+			"       FORMAT := { raw | preallocated }\n"
+			);
 }
 
 static int plooptool_convert(int argc, char **argv)
@@ -801,17 +795,10 @@ static int plooptool_convert(int argc, char **argv)
 	struct ploop_disk_images_data *di;
 	int mode = -1;
 
-	while ((i = getopt(argc, argv, "t:")) != EOF) {
+	while ((i = getopt(argc, argv, "f:")) != EOF) {
 		switch (i) {
-		case 't':
-			if (!strcmp(optarg, "raw"))
-				mode = PLOOP_RAW_MODE;
-			else if (!strcmp(optarg, "preallocated"))
-				mode = PLOOP_EXPANDED_PREALLOCATED_MODE;
-			else {
-				usage_convert();
-				return -1;
-			}
+		case 'f':
+			mode = parse_format_opt(optarg);
 			break;
 		default:
 			usage_convert();
@@ -841,7 +828,7 @@ static int plooptool_convert(int argc, char **argv)
 
 static void usage_info(void)
 {
-	fprintf(stderr, "Usage: ploop info\n");
+	fprintf(stderr, "Usage: ploop info DiskDescriptor.xml\n");
 }
 
 static void print_info(struct ploop_info *info)
